from configparser import ConfigParser, BasicInterpolation
import os
from typing import List
import logging
import torch


class ConfigLoader:
    __speech2textmodel = "speech-to-text"
    __sentenceimportancemodel = "sentence-importance"
    # add model here

    _supported_language = ["english"]

    def __init__(self, configpath: str):

        self._loadconfigfile(configpath)

    def _loadconfigfile(self, configpath: str):

        if os.path.exists(configpath) is False:

            logging.error(f"Config file does not exist: {configpath}")
            return

        elif configpath.endswith(".ini") is False:

            logging.error(f"Config file does not exist: {configpath}")
            return

        self._configpath = configpath

        self.config = ConfigParser(interpolation=BasicInterpolation())
        self.config.read(configpath)

        self._rootpath = self.config["DEFAULT"]["rootpath"]
        self._language = self.config["DEFAULT"]['language']

        self.islanguagesupported(self._language)

    def islanguagesupported(self, language: str):

        issupported = True if language in self._supported_language else False

        if issupported is False:

            logging.error(
                f"Specified language {language} is not supported")

        return issupported

    def reloadconfigfile(self, configpath: str = None):
        """reload config file if necessary"""
        self._loadconfigfile(
            configpath if configpath is not None else self._configpath)

    @property
    def language(self):
        return self._language

    @property
    def supportedlanguages(self) -> List[str]:

        return self._supported_language

    @property
    def speech2textmodelpath(self):

        return self.config[self._language][self.__speech2textmodel]

    @property
    def correctspelling(self):
        return self.config[self._language].getboolean('correctspelling')

    @property
    def sentenceimportancemodelpath(self):

        return self.config[self._language][self.__sentenceimportancemodel]

    def usecuda(self) -> str:
        """
        To use "cuda" or "cpu" - for both model, processor/tokenizer and input
        """

        return "cuda" if self.config["DEFAULT"].getboolean("usecuda") is True else "cpu"

    @language.setter
    def language(self, inputlanguage: str):
        """
        allow the setting of language
        """

        if self.islanguagesupported(inputlanguage) is False:

            logging.warning("Changing of language failed")

        self._language = inputlanguage



import torch
from conversationalnlp.models.sentenceimportance.modelloader import ModelLoader


class SentenceImportancePredict:

    def __init__(self, modelloader: ModelLoader):

        self._modelloader = modelloader
        self._id2label = self._modelloader.model.config.id2label

    def classify(self, text: str) -> str:

        inputs = self._modelloader.tokenizer(text, padding=True,
                                             truncation=True, return_tensors="pt")

        if self._modelloader.device == "cuda":
            inputs = inputs.to("cuda")

        outputs = self._modelloader.model(**inputs)

        prediction = torch.nn.functional.softmax(outputs.logits, dim=1)
        final_prediction = torch.argmax(prediction, dim=1)

        return self._id2label[final_prediction[0].item()]



from conversationalnlp.utils import ConfigLoader
from transformers import Wav2Vec2Processor, AutoModelForCTC
import torch
import logging


class ModelLoader:

    __device = "cpu"

    def __init__(self, configloader: ConfigLoader):

        speech2text_model = configloader.speech2textmodelpath

        self.__correctspelling = configloader.correctspelling

        self.__processor = Wav2Vec2Processor.from_pretrained(
            speech2text_model, use_auth_token=True)

        self.__model = AutoModelForCTC.from_pretrained(
            speech2text_model, use_auth_token=True)

        self.__setdevice(configloader.usecuda())

    def __init__(self, model: AutoModelForCTC, processor: Wav2Vec2Processor, correctspelling=True, device="cuda"):

        self.__processor = processor
        self.__model = model

        self.__correctspelling = correctspelling

        self.__setdevice(device)

    @property
    def processor(self):

        return self.__processor

    @property
    def model(self):

        return self.__model

    @property
    def device(self):
        return self.__device

    @property
    def tocorrectspelling(self):

        return self.__correctspelling

    def __setdevice(self, device):

        if device != "cuda" and device != "cpu":

            logging.error(
                f"Setting of device failed with param not either cpu or cuda. Param set {device}")

        if device == "cuda" and torch.cuda.is_available():

            self.__model.cuda()
            self.__device = "cuda"

        elif device == "cuda" and torch.cuda.is_available() is False:

            logging.warning(
                "Config set to cuda is true but cuda not found in system. Auto set to cpu.")



from conversationalnlp.utils import customdatetime
import pandas as pd
import platform
import logging
import os
from typing import List


class ModelTracker:
    """
    Save to modeltracker_consolidation.xlsx with two sheets

    First Sheet - Training Log: {os_name}_{date}_progress

    | wer | datetime | 
    |  ---------------------  | ---------------------  |
    | 0.70 | 2022Mar04_19-45-18 |
    | 0.2 | 2022Mar04_19-45-18 |

    Second sheet - Training Summary: {os_name}_{date}_summary

    No exact format as long as writable

    | Description | Content | 
    |  ---------------------  | ---------------------  |
    | Base Model | wav2vec2-base |
    | Training Shape | (1234, 5) |
    | Training Shape | (1234, 5) |
    | Dataset | file1 | 
    """

    columndict = dict()
    modelcolumnname = "model"
    datecolumnname = "datetime"

    def __init__(self, metrics: List[str], trackerpath: str, modelsummary: pd.DataFrame = None):
        """
        Create attributes to track models performances

        Attributes
        ----------
        metrics_name : str 
            Name of metric to track. 
        trackerpath : str 
            Path of output excel file <relativepath>\test.xlsx
        trackerfileheader : str, optional 
            Filename header

        Returns
        -------
        None
        """

        if trackerpath.endswith(".xlsx") is False and os.path.exists(trackerpath) is False:

            logging.error(
                "Model tracker trackerpath is neither a valid path nor a valid path. Setting modeltracker failed.")
            return

        elif trackerpath.endswith(".xlsx") is False and os.path.exists(trackerpath) is True:

            trackerpath = os.path.join(trackerpath, "modeltracker.xlsx")
            logging.info(f"Model tracker set at path {trackerpath}")

        column_name = [self.modelcolumnname,
                       self.datecolumnname]  # default column name

        column_name.extend(metrics)

        for name in column_name:

            self.columndict[name] = []

        self.trackerpath = trackerpath

        # Append date time to prevent duplication
        datetime = customdatetime.getstringdatetime()
        self.trackersheetbase = platform.system().lower() + "_" + datetime.lower()

        logging.info(f"Model tracker: {self.trackerpath}")

        if modelsummary is not None:
            self._save_to_model_summary_file(modelsummary)

        self._save_to_model_progress_file()

    def additem(self, info: dict):
        """
        Create attributes to track models performances

        Attributes
        ----------
        metrics_name : str 
            Name of metircs to track. Each will become a column

        Returns
        -------
        None
        """
        info[self.datecolumnname] = customdatetime.getstringdatetime()

        if self.modelcolumnname not in info.keys():

            info[self.modelcolumnname] = None

        for key, value in info.items():

            if key not in self.columndict.keys():
                logging.info(
                    f"New {key} key not available in modeltracker. Value not saved.")
            else:
                self.columndict[key].append(value)

        keysmissing = list(set(self.columndict.keys()) - set(info.keys()))

        if keysmissing:
            logging.warning(
                f"Keys necessary for model tracking missing. {keysmissing}")

        # to prevent column name not same
        for key in keysmissing:

            self.columndict[key].append(None)

        # save every iteration
        self._save_to_model_progress_file()

    def _save_to_model_progress_file(self):
        """
        Write to output csv file
        Attributes
        ----------
        None

        Returns
        -------
        None
        """
        df = pd.DataFrame(self.columndict)

        mode = 'a' if os.path.exists(self.trackerpath) is True else 'w'

        with pd.ExcelWriter(self.trackerpath, mode=mode) as writer:

            sheet_name = self.trackersheetbase + "_1"

            workBook = writer.book
            try:
                workBook.remove(workBook[sheet_name])
            except:
                pass
            finally:
                df.to_excel(writer, sheet_name=sheet_name, index=False)
                writer.save()

    def _save_to_model_summary_file(self, df):
        """
        Write to output csv file
        Attributes
        ----------
        None

        Returns
        -------
        None
        """

        mode = 'a' if os.path.exists(self.trackerpath) is True else 'w'

        with pd.ExcelWriter(self.trackerpath, mode=mode) as writer:

            sheet_name = self.trackersheetbase + "_0"

            workBook = writer.book
            try:
                workBook.remove(workBook[sheet_name])
            except:
                pass
            finally:
                df.to_excel(writer, sheet_name=sheet_name, index=False)
                writer.save()



import pandas as pd
import random


def show_random_elements(dataset, num_examples=1):
    assert num_examples <= len(
        dataset), "Can't pick more elements than there are in the dataset."
    picks = []
    for _ in range(num_examples):
        pick = random.randint(0, len(dataset)-1)
        while pick in picks:
            pick = random.randint(0, len(dataset)-1)
        picks.append(pick)

    df = pd.DataFrame(dataset[picks])

    from IPython.display import display, HTML
    display(HTML(df.to_html()))




from conversationalnlp.models.wav2vec2.dataset import Wav2Vec2Dataset
from conversationalnlp.models.wav2vec2.modelloader import ModelLoader
from autocorrect import Speller
from typing import Dict, List
import torch


class Wav2Vec2Predict:

    _wav2vec2dataset = Wav2Vec2Dataset()
    _spellingcorrection = Speller()

    # text from speech2 text model
    __predictedtext_param = "predicted_text"

    # text from spelling correction
    __correctedtext_param = "corrected_text"

    def __init__(self, modelloader: ModelLoader):

        self._modelloader = modelloader

    def predictfiles(self, audiofilepaths: List[str]) -> List[Dict[str, str]]:
        """
        Predict single file
        Return Dict[str, str]
        {
            "text": ["helo world", "hi"]
            "corrected_text": ["hello world", "hi"]
        }
        """
        textdict = {self.__predictedtext_param: [],
                    self.__correctedtext_param: []}

        iterator = map(self.predictfile, audiofilepaths)

        for unit in iterator:
            textdict[self.__predictedtext_param].append(
                unit[self.__predictedtext_param])
            textdict[self.__correctedtext_param].append(
                unit[self.__correctedtext_param])

        return textdict

    def predictfile(self, audiofilepath: str) -> Dict[str, str]:
        """
        Predict single file
        Return Dict[str, str]
        {
            "text": "helo world",
            "corrected_text": "hello world"
        }
        """
        tensor_audio = self._wav2vec2dataset.vectorizefile2tensor(
            self._modelloader, audiofilepath)

        return self.predict(tensor_audio)

    def predict(self, tensor_audio: torch.tensor) -> Dict[str, str]:

        with torch.no_grad():
            logits = self._modelloader.model(tensor_audio).logits

            pred_ids = torch.argmax(logits, dim=-1)

            text = self._modelloader.processor.batch_decode(pred_ids)[0]

            corrected_text = self._spellingcorrection(
                text) if self._modelloader.tocorrectspelling is True else None

            return {self.__predictedtext_param: text, self.__correctedtext_param: corrected_text}



import numpy as np
import pandas as pd
import os

class LabellingFunc:

    def __init__(self):

        keywordspath = os.path.join(os.path.dirname(os.path.abspath(__file__)), "metadata", "keywords0.txt")
        filein = open(keywordspath, "r")
        raw_keywords_class0 = filein.readlines()
        self.keywords_class0 = list(map(str.strip, raw_keywords_class0))

        self._labelidxlut =  {"important": 1, "not important": 0}
        self._labelstrlut = {v: k for k, v in self._labelidxlut.items()}

    @property
    def labelidxlut(self):

        return self._labelidxlut

    @property
    def labelstrlut(self):

        return self._labelstrlut

    def filter(self, sentence) -> list:
        """
        - Remove words that matches keywords of class 0
        - Remove repetitive words
        """
        raw_words = sentence.split()
        processed_words = [str.lower(word.replace('\'', '')) for word in raw_words]

        repetitives = pd.value_counts(np.array(processed_words))

        for word, count in repetitives.items():
            for _ in range(0, count - 1):
                processed_words.remove(word)

        return [word for word in processed_words if word not in self.keywords_class0]

    def pseudolabel(self, words :str) -> int:
        """
        Label a sequence/sentence with importance classification based on heuristics
        
        Attributes
        ----------
        sentence : str
            a single sentence

        Returns
        -------
        int
        0: not important
        1: important
        -1: needs further inspection
        """
        wordsfiltered = self.filter(words)

        wordcount = len(wordsfiltered)

        if wordcount < 3:

            if (wordcount == 1) and (words.lower() in ["okay", "ok", "yes", "no"]):
                return 1
            else:
                return 0
        elif wordcount > 5:
            return 1
        else:
            return -1
        
        

        


import torch

from transformers import Wav2Vec2Processor
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Union

@dataclass
class DataCollatorCTCWithPadding:
    """
    Data collator that will dynamically pad the inputs received.
    Args:
        processor (:class:`~transformers.Wav2Vec2Processor`)
            The processor used for proccessing the data.
        padding (:obj:`bool`, :obj:`str` or :class:`~transformers.tokenization_utils_base.PaddingStrategy`, `optional`, defaults to :obj:`True`):
            Select a strategy to pad the returned sequences (according to the model's padding side and padding index)
            among:
            * :obj:`True` or :obj:`'longest'`: Pad to the longest sequence in the batch (or no padding if only a single
              sequence if provided).
            * :obj:`'max_length'`: Pad to a maximum length specified with the argument :obj:`max_length` or to the
              maximum acceptable input length for the model if that argument is not provided.
            * :obj:`False` or :obj:`'do_not_pad'` (default): No padding (i.e., can output a batch with sequences of
              different lengths).
        max_length (:obj:`int`, `optional`):
            Maximum length of the ``input_values`` of the returned list and optionally padding length (see above).
        max_length_labels (:obj:`int`, `optional`):
            Maximum length of the ``labels`` returned list and optionally padding length (see above).
        pad_to_multiple_of (:obj:`int`, `optional`):
            If set will pad the sequence to a multiple of the provided value.
            This is especially useful to enable the use of Tensor Cores on NVIDIA hardware with compute capability >=
            7.5 (Volta).
    """

    processor: Wav2Vec2Processor
    padding: Union[bool, str] = True
    max_length: Optional[int] = None
    max_length_labels: Optional[int] = None
    pad_to_multiple_of: Optional[int] = None
    pad_to_multiple_of_labels: Optional[int] = None

    def __call__(self, features: List[Dict[str, Union[List[int], torch.Tensor]]]) -> Dict[str, torch.Tensor]:
        # split inputs and labels since they have to be of different lenghts and need
        # different padding methods
        input_features = [{"input_values": feature["input_values"]} for feature in features]
        label_features = [{"input_ids": feature["labels"]} for feature in features]

        batch = self.processor.pad(
            input_features,
            padding=self.padding,
            max_length=self.max_length,
            pad_to_multiple_of=self.pad_to_multiple_of,
            return_tensors="pt",
        )
        with self.processor.as_target_processor():
            labels_batch = self.processor.pad(
                label_features,
                padding=self.padding,
                max_length=self.max_length_labels,
                pad_to_multiple_of=self.pad_to_multiple_of_labels,
                return_tensors="pt",
            )

        # replace padding with -100 to ignore loss correctly
        labels = labels_batch["input_ids"].masked_fill(labels_batch.attention_mask.ne(1), -100)

        batch["labels"] = labels

        return batch


import os
import urllib
import zipfile
import string
import random
import shutil
import hashlib
from . import customdatetime
import logging
import sys

logging.basicConfig(
            format="%(asctime)s | %(levelname)s | %(name)s |  [%(filename)s:%(lineno)d] %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
            level=os.environ.get("LOGLEVEL", "INFO").upper(),
            stream=sys.stdout,
        )

def checksum(path): 
    f = open(path, 'rb')
    data = f.read()
    hash = hashlib.md5(data).hexdigest().lower()
    return hash

def downloadurl(zipinfo) -> bool:
    """
    Download url to local file

    Attributes
    ----------
    Dict of 
        url : str 
            URL to download
        filepath : str 
            File path to save to 
        hash : str
            Hash of file

    Returns
    -------
    Boolean if zip successfully download/exist
    """
    hash =  zipinfo['hash'].lower()

    if os.path.exists(zipinfo['filepath']) and hash is not None:

        if checksum(zipinfo['filepath']) == hash:

            logging.info(f"Checksum same. Skip in downloading {zipinfo['filepath']}")  
            return True

    try:
        _, _ = urllib.request.urlretrieve(zipinfo['url'], zipinfo['filepath'])
        logging.info(f"File downloaded at {zipinfo['filepath']}")

        
    except urllib.error.HTTPError as err:
        
        logging.error(f"Error in downloading {zipinfo['url']}: {err}")
        return False

    return True

def unzip(unzipinfo):
    """
    Unzip local zip path to uncompressed file

    Attributes
    ----------
    dict of 
        filepath : str 
            Path of zip file
        unzippath : str
            Folder to unzip to 
        overwrite: bool, optional 
            Existing folder will be deleted

    Returns
    -------
    None
    """
    unzippath = unzipinfo['unzippath']
    
    if os.path.exists(unzippath):

        if(('overwrite' in unzipinfo.keys()) and (unzipinfo['overwrite'] is True)):

            shutil.rmtree(unzippath)
            
        else:
            
            logging.info(f"{unzippath} exist. Unzipping skipped")
            return

    os.mkdir(unzippath)

    tempfolder = ''.join(random.choice(string.ascii_lowercase) for i in range(8))
    temppath = os.path.join(unzippath, tempfolder)

    with zipfile.ZipFile(unzipinfo['filepath'], 'r') as zipref:
        zipref.extractall(temppath)

    iterfiles = os.walk(temppath)
    for item in iterfiles:

        #assume only have one level sub directory
        currentfolder = item[0].split(os.sep)[-1]

        if(currentfolder != tempfolder):
            currentunzippath = os.path.join(unzippath, currentfolder)
            os.mkdir(currentunzippath)

        else:
            currentunzippath = unzippath

        for file in item[2]:
            src = os.path.join(item[0], file)
            dest = os.path.join(currentunzippath, file)
            shutil.move(src, dest)

    #remove temp path
    shutil.rmtree(temppath)
    
        
def createfolders(path):

    if not os.path.exists(path):

        logging.info(f"Folder created: {path}")
        os.makedirs(path)
    else:
        logging.info(f"Folder exist: {path}")


def generatefoldername(headername = ""):

    """
    Create unique folder name based on date

    Attributes
    ----------
    filepath : str 
        Path of zip file

    Returns
    -------
    foldername : str
        Unique folder name
    """

    strdatetime = customdatetime.getstringdatetime()

    return (headername + "_" + strdatetime) if headername != "" else strdatetime



import torch
from conversationalnlp.utils import ConfigLoader
from transformers import AutoTokenizer, AutoModelForSequenceClassification


class ModelLoader:

    # TODO: Does the checkpoint potentially needs to change in the future?
    tokenizer_checkpoint = "bert-base-uncased"
    self.device = "cpu"

    def __init__(self, configloader: ConfigLoader):

        sentenceimportance_model = configloader.sentenceimportancemodelpath

        self._correctspelling = configloader.correctspelling

        self.__tokenizer = AutoTokenizer.from_pretrained(
            self.tokenizer_checkpoint)

        self.__model = AutoModelForSequenceClassification.from_pretrained(
            sentenceimportance_model, use_auth_token=True)

        self.__device = configloader.usecuda()
        if self.device == "cuda":
            self.__model.cuda()

    def __init__(self, model: AutoModelForSequenceClassification, tokenizer: AutoTokenizer, tocuda=True):
        """
        device: when not set, will set to cuda if exist
        """

        self.__model = model
        self.tokenizer = tokenizer

        if tocuda is True and torch.cuda.is_available():
            self.__device = "cuda"
            self.__model.cuda()

    @property
    def tokenizer(self):
        return self.__tokenizer

    @property
    def model(self):

        return self.__model

    @property
    def device(self):
        return self.__device



import os
import logging
import platform
import shutil
import pandas as pd
from typing import List
from conversationalnlp.models.wav2vec2.datasanitycheck import Wav2Vec2DataSanityCheck
from conversationalnlp.utils import filesys


class Wav2Vec2FilePrep:
    """
    Generate wav2vec2_train.csv & wav2vec2_test.csv in <datapath>/wav2vec2compilation
    """

    def __init__(self):

        self.trainsplit = 0.8
        # TODO: change this into parameters that can toggle if necessary

        self.__sanitycheck = Wav2Vec2DataSanityCheck()

        self.__wav2vec2columns = ["file", "text"]

    def run(self, fileinfo) -> dict:
        """
        Prepare train and test input file to train wav2vec2

        Attributes
        ----------
        Dict of 
            wav2vec2datapath : str 
                Root path containing audio, audio-chunks, text folder
            os: str
                Operating system to decide on file separation
            overwrite : bool, optional
                Whether to overwrite output file if exist
            textfilelist : list 
                List of csv text paths to be included.
                This is to selectively include / exclude certain text dat a

        Returns
        -------
        Dict of 
            train: str
                Absolute file path to wav2vec2 train file

            test: str
                Absolute file path to wav2vec2 test file

        None
            If process abort
        """

        textpath = os.path.join(fileinfo['wav2vec2datapath'], "text")

        # check if input folder exist
        if os.path.exists(textpath) is False:
            logging.error(
                f"Input folder f{textpath} to generate train and test data not exist. Operation aborted.")
            return None

        audiochunkpath = os.path.join(
            fileinfo['wav2vec2datapath'], "audio-chunks")
        wav2vec2compilationpath = os.path.join(
            fileinfo['wav2vec2datapath'], "wav2vec2compilation")

        trainfilepath = os.path.join(
            wav2vec2compilationpath, "wav2vec2_train.csv")
        testfilepath = os.path.join(
            wav2vec2compilationpath, "wav2vec2_test.csv")
        faultyfilepath = os.path.join(
            wav2vec2compilationpath, "wav2vec2_faulty.csv")

        # create/overwrite output folder
        if(("overwrite" in fileinfo.keys()) and (fileinfo['overwrite'] is True) and (os.path.exists(wav2vec2compilationpath))):
            logging.info(f"Folder of {wav2vec2compilationpath} deleted")
            shutil.rmtree(wav2vec2compilationpath)

        if os.path.exists(wav2vec2compilationpath) is False:
            filesys.createfolders(wav2vec2compilationpath)

        # append path (in another function)

        textfilelist = [textpath + os.sep +
                        i for i in fileinfo['textfilelist']]

        df = pd.DataFrame()
        faultydf = pd.DataFrame()

        for filepath in textfilelist:

            logging.info(f"Process data from {filepath}")

            currentdf = pd.read_csv(filepath, engine='python')

            # fix separator according to OS
            currentdf['file'] = self._alignOSpath(currentdf['file'].tolist())

            # add proper path
            currentdf['file'] = audiochunkpath + os.sep + currentdf['file']

            # sanity check files if not exist
            rawcurrentdfshape = currentdf.shape
            sanitylut = self.__sanitycheck.run(
                currentdf, textcolumn="text", filepathcolumn="file")

            currenttrimdf = sanitylut['clean']
            currentfaultydf = sanitylut['faulty']

            currenttrimdf = currenttrimdf.dropna(
                axis=1)  # drop nan rows if exist
            trimdfshape = currenttrimdf.shape

            if trimdfshape[0] != rawcurrentdfshape[0]:
                logging.warning(
                    f"Dropped {rawcurrentdfshape[0] - trimdfshape[0]} invalid rows from file {filepath}")

            if currentfaultydf.empty is False:

                faultydf = pd.concat(
                    [faultydf, currentfaultydf], axis=0, ignore_index=True)

            df = pd.concat([df, currenttrimdf], axis=0, ignore_index=True)

        if df.empty is True:

            logging.error(
                "DataFrame empty. Train and test file cannot be generated.")
            return None

        # filter dataframe for wav2vec2 input
        df = df[self.__wav2vec2columns]

        # shuffle data
        df_train, df_test = self._shuffleandsplitdata(df)

        # save to train and test data file
        df_train.to_csv(trainfilepath, index=False)
        df_test.to_csv(testfilepath, index=False)
        faultydf.to_csv(faultyfilepath, index=False)

        logging.info(f"Training shape: {df_train.shape}")
        logging.info(f"Testing shape: {df_test.shape}")
        logging.info(f"Faulty shape: {faultydf.shape}")

        return {"train": trainfilepath, "test": testfilepath, "faulty": faultyfilepath}

    def _alignOSpath(self, filelist: List[str]) -> List[str]:
        """
        Correct file separator according to operating sytem

        Attributes
        ----------

        Returns
        -------
        List[str]
        """
        OS = platform.system()

        if OS == "Linux":

            alignedfilelist = [i.replace("\\", os.sep) for i in filelist]
        else:
            alignedfilelist = filelist

        return alignedfilelist

    def _shuffleandsplitdata(self, df) -> set:
        """
        Shuffle and split dataframe

        Attributes
        ----------
        df : DataFrame 
            input dataframe

        Returns
        -------
        Set of two dataframe. One for train purpose, another for test purpose
        """

        df = df.sample(frac=1).reset_index(drop=True)

        # split into 8:2
        train_rows = int(df.shape[0] * self.trainsplit)

        df_train = df.iloc[0:train_rows, :]
        df_test = df.iloc[train_rows:df.shape[0], :]

        return (df_train, df_test)



from conversationalnlp.models.wav2vec2.modelloader import ModelLoader
from datasets import load_dataset
from datasets.dataset_dict import DatasetDict
from transformers import Wav2Vec2Processor
from conversationalnlp.utils import display
import soundfile as sf
import logging
import re
import os
import json
import torch
"""
Dataset to train wav2vec2 expected to be in the following format 

| text | file |
| ---------------------  | ---------------------  |
| Hello world | audiofolder/chunk1.flac |
| Yes this is a sample text | audiofolder/chunk2.flac |

"""


class Wav2Vec2Dataset:
    """
    Long input sequences require a lot of memory. 
    Since Wav2Vec2 is based on self-attention the memory requirement scales quadratically with the input length for long input sequences
    TODO: remove rows with audio length > n seconds/word count > n

    """

    chars_to_ignore_regex = '[\,\?\.\!\-\;\:\"]'

    def __init__(self):
        pass

    def loadHFdataset(self, dataset: DatasetDict):
        """
        Attributes
        ----------
        dataset : datasets.dataset_dict.DatasetDict
            In house dataset hosted on Hugging face hub
        """

        self.dataset = self._normalizetext(dataset)
        self.dataset_shape = self.dataset.shape

        display.show_random_elements(self.dataset["train"])

    def loadcustomdataset(self, datadict: dict):
        """
        Attributes
        ----------
        datadict : dict

            dict with keys "train" and "test", both stating absolute csv file paths
        """
        rawdataset = load_dataset("csv", data_files=datadict)
        display.show_random_elements(rawdataset["train"])

        self.dataset = self._normalizetext(rawdataset)
        self.dataset_shape = self.dataset.shape

        display.show_random_elements(self.dataset["train"])

    def getshape(self) -> dict():
        """
        Return shape of train and test data shape
        """

        return self.dataset_shape

    def _prepare_dataset(self, batch):
        """
        - Process the dataset to the format expected by the model for training
        - Use map(...) function
        1. Load and resample the audio data, simply by calling batch['audio']. 
            - If batch['audio'] not exist, will fall back to read from file 
        2. Extract the input_values from the loaded audio file (In our case, the Wav2Vec2Processor only normalized the data)
        3. Encode the transcriptions to label ids
        """
        audio = None
        samplerate = None

        if "audio" in batch.features.keys():
            audio = batch["audio"]["array"]
            samplerate = batch["audio"]["sampling_rate"]

        else:
            # use file
            if(os.path.exists(batch['file']) != True):

                logging.error(f"{batch['file']} not available")

            else:
                audio, samplerate = sf.read(batch['file'])

        # batched output is "un-batched" to ensure mapping is correct
        batch["input_values"] = self.processor(
            audio, sampling_rate=samplerate).input_values[0]

        with self.processor.as_target_processor():
            batch["labels"] = self.processor(batch["text"]).input_ids

        return batch

    def vectorize(self, processor: Wav2Vec2Processor, datacolumn="train") -> DatasetDict:
        """
        Prepare dataset with def prepare_dataset
        Attributes
        ----------
        processor : Wav2Vec2Processor
            Wav2Vec2Processor
        datacolumn : str, optional
            Either "train" or "test" for the column, default as train

        Returns
        -------
        datasets.dataset_dict.DatasetDict
        """
        self.processor = processor

        dataset = self.dataset.map(
            self._prepare_dataset, remove_columns=self.dataset.column_names[datacolumn])  # , num_proc=4)

        del self.dataset

        return dataset

    def vectorizefile2tensor(self, modelloader: ModelLoader, audiopath: str) -> torch.tensor:
        """
        Prepare dataset with def prepare_dataset
        Attributes
        ----------
        processor : Wav2Vec2Processor
            Wav2Vec2Processor
        audiopath : str
            File path to audio file (.wav/flac)

        Returns
        -------
        datasets.dataset_dict.DatasetDict
        """
        if(os.path.exists(audiopath) != True):

            logging.error(f"{audiopath} not available")

            return None

        else:
            audio, samplerate = sf.read(audiopath)

            np_audio = modelloader.processor(
                audio, sampling_rate=samplerate).input_values[0]

            tensor_audio = torch.tensor(
                [np_audio.tolist()], device=modelloader.device)

            return tensor_audio

    def _normalizetext(self, dataset: DatasetDict) -> DatasetDict:
        """
        Normalize text 

        Returns
        -------
        datasets.dataset_dict.DatasetDict

        """
        logging.info("Normalize text")
        return dataset.map(self._remove_special_characters)

    def _remove_special_characters(self, batch):
        """
        Remove special characters from transcriptions and normalized them to lower-case only
        """
        batch["text"] = re.sub(self.chars_to_ignore_regex,
                               '', batch["text"]).lower()
        return batch

    def buildvocabulary(self, vocabpath: str, vocabfilename="vocab.json") -> str:
        """
        Build vocabulary from all the unique characters

        Attributes
        ----------
        vocabpath : str
            Path to save vocabulary file
        vocabfilename : str, optional
            File name of vocabulary file

        Returns
        -------
        vocababspath : str
            absolute path of vocabulary file         
        """
        logging.info("Build vocabulary from all unique characters")

        vocabs = self.dataset.map(self._extract_all_chars, batched=True, batch_size=-1,
                                  keep_in_memory=True, remove_columns=self.dataset.column_names["train"])

        vocab_list = list(set(vocabs["train"]["vocab"][0]) | set(
            vocabs["test"]["vocab"][0]))
        vocab_dict = {v: k for k, v in enumerate(vocab_list)}

        # space (" ") has its own token class, replace with more visible character
        vocab_dict["|"] = vocab_dict[" "]
        del vocab_dict[" "]

        # add an "unknown" token so that model can later deal with characters not encountered
        vocab_dict["[UNK]"] = len(vocab_dict)

        # Add padding token that corresponds to CTC's "blank token"
        # The "blank token" is a core component of the CTC algorithm
        vocab_dict["[PAD]"] = len(vocab_dict)

        logging.info(f"Vocabulary size: {len(vocab_dict)}")
        logging.info(
            f"The pretrained Wav2Vec2 checkpoint will have an output dimension of {len(vocab_dict)}")

        vocababspath = os.path.join(vocabpath, vocabfilename)

        with open(vocababspath, 'w') as vocabfile:
            json.dump(vocab_dict, vocabfile)

        return vocababspath

    def _extract_all_chars(self, batch):
        """
        Process input string to get unique characters
        """
        all_text = " ".join(batch["text"])
        vocab = list(set(all_text))
        return {"vocab": [vocab], "all_text": [all_text]}



import pandas as pd
import logging
import os
from typing import Dict


class Wav2Vec2DataSanityCheck:
    """
    To validate the source data prior to training of Wav2Vec2
    """

    __cleandfparam = "clean"
    __faultydfparam = "faulty"

    def run(self, df, textcolumn: str = "text", filepathcolumn: str = "file") -> Dict[str, pd.DataFrame]:
        """
        - Validate dataset of column text where vocabulary fits in vocabulary size
        - Validate file path if exist
        Attributes
        ----------
        textcolumn : str, optional
            Column name for text 
        filepathcolumn : str, optional
            Column name for filepath

        Returns
        -------
        dict of dataframe with "clean" and "faulty"

        Dataframe with errors of following columns
        index : index in original dataframe 
        textcolumn param : text 
        filepathcolumn param: filepath 
        note: whether is "path not exist" or "digit not valid"

        """
        faultyindexlist = []
        faultytextcolumnlist = []
        faultyfilepathcolumnlist = []
        faultynotelist = []  # "path not exist", "digit not valid"

        if df.empty is True:
            logging.warning("Input dataframe for sanity check is empty")
            return None

        # Remove nan column

        dforishape = df.shape
        df = df.dropna(axis=1)

        if df.shape[0] != dforishape[0]:

            logging.info(
                f"Dataframe shape not same, contains invalid rows. Removed rows {dforishape[0] - df.shape[0]}")

        columns_missing = list(
            filter(lambda column: column not in df.columns, [textcolumn, filepathcolumn]))

        # check if columns required is missing
        if columns_missing:
            logging.error(f"Columns required {columns_missing} missing")
            return None

        for index, row in df.iterrows():

            isfaulty = False

            if any(chr.isdigit() for chr in row[textcolumn]) is True:
                isfaulty = True
                faultynotelist.append("digit not valid")

            elif os.path.exists(row[filepathcolumn]) is False:
                isfaulty = True
                faultynotelist.append("path not exist")

            if isfaulty:
                # 1 for header, 1 for table start with index 1 than index 0
                faultyindexlist.append(index)
                faultytextcolumnlist.append(row[textcolumn])
                faultyfilepathcolumnlist.append(row[filepathcolumn])

        # no faulty data
        if not faultyindexlist:
            logging.info("No faulty lines")
            return {self.__cleandfparam: df, self.__faultydfparam: pd.DataFrame()}

        dffaulty = pd.DataFrame({"index": faultyindexlist, textcolumn: faultytextcolumnlist,
                                 filepathcolumn: faultyfilepathcolumnlist, "note": faultynotelist})

        dfcleaned = df.drop(faultyindexlist).reset_index(drop=True)

        return {self.__cleandfparam: dfcleaned, self.__faultydfparam: dffaulty}



from conversationalnlp.utils import ModelTracker
from transformers import Wav2Vec2Processor
from datasets import load_metric
from typing import List
import numpy as np


class WER:
    """
    https://github.com/huggingface/datasets/blob/master/metrics/wer/wer.pyz
    """
    modeltracker = None

    def __init__(self, processor: Wav2Vec2Processor = None, modeltracker: ModelTracker = None):
        """
        pass modeltracker to track wer metrics of each iteration
        """
        self.__wer_metric = load_metric(self.__str__())
        # processor is only needed for training, not necessary for plain computation of wer score
        self.__processor = processor
        self.modeltracker = modeltracker

    def evaluate(self, groundtruthtext: List[str], predictedtext: List[str]) -> float:
        """
        used in evaluation between two pairs of text to compute wer score
        """
        groundtruthtext = list(map(lambda x: x.lower(), groundtruthtext))
        predictedtext = list(map(lambda x: x.lower(), predictedtext))

        wer = self.__wer_metric.compute(
            predictions=predictedtext, references=groundtruthtext)

        if self.modeltracker is not None:
            self.modeltracker.additem({self.__str__(): wer})

        return wer

    def train(self, pred) -> dict:
        """
        Use in training 
        """
        pred_logits = pred.predictions
        pred_ids = np.argmax(pred_logits, axis=-1)

        pred.label_ids[pred.label_ids == -
                       100] = self.__processor.tokenizer.pad_token_id

        pred_str = self.__processor.batch_decode(pred_ids)

        # we do not want to group tokens when computing the metrics
        label_str = self.__processor.batch_decode(
            pred.label_ids, group_tokens=False)

        wer = self.__wer_metric.compute(
            predictions=pred_str, references=label_str)

        werdict = {self.__str__(): wer}

        if self.modeltracker is not None:

            self.modeltracker.additem(werdict)

        return werdict

    def __str__(self):

        return "wer"



import re

def remove_emoji(instr: str) -> str:
    emoji_pattern = re.compile("["
        u"\U0001F600-\U0001F64F"  # emoticons
        u"\U0001F300-\U0001F5FF"  # symbols & pictographs
        u"\U0001F680-\U0001F6FF"  # transport & map symbols
        u"\U0001F1E0-\U0001F1FF"  # flags (iOS)
                           "]+", flags=re.UNICODE)
    return emoji_pattern.sub(r'', instr)

def cleansentence(instr: str) -> str:
    """
    Clean sentence from 
    - strip white space
    - \n \t
    - remove emoji
    """    
    
    clean_list = ["tldr:", "\n", "\t"]
    
    for clean_item in clean_list:
        instr = instr.replace(clean_item, "")
       
    #strip white space
    workingstr = instr.strip()
        
    outstr = remove_emoji(workingstr)
        
    return outstr.lower()

def splitbysentence(instr: str) -> list:
    """
    split a string and return list of string following separator \".\", \".\", \".\"
    Note: Current version exclude the separator
    
    #splitbysentence("Hello. My name is C? But! Really, It's impossible.")
    """
    
    workinglist = [instr]
    
    splitter = [".", "?", "!"]
    
    for current_split in splitter:
        
        outlist = []
        for sentence in workinglist:
            
            templist = list(filter(lambda x : len(x) > 1, sentence.split(current_split)))
            outlist.extend(templist)
            
        workinglist = outlist
        
    workinglist = [cleansentence(i) for i in workinglist]
        
    outlist = list(filter(lambda instr: len(instr) > 1, workinglist))
                
    return outlist


#teststr = 'Hi! Howdy! 😎.\ntldr: i am recovering alcoholic'
#splitbysentence(teststr)


from setuptools import find_packages
from setuptools import setup

with open("readme.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name='conversationalnlp',
    version='0.0.3post4',
    install_requires=['torch==1.10.0', 'torchaudio==0.10.0', 'transformers==4.20.1',
                      'autocorrect==2.6.1', 'SoundFile==0.10.3.post1'],
    include_package_data=True,
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    python_requires=">=3.8.5",
    url='https://github.com/codenamewei/conversationalnlp',
    license='MIT',
    author='codenamewei',
    author_email='codenamewei@gmail.com',
    description='Your main project',
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)



import numpy as np
import pandas as pd
import os

class LabellingFunc:

    def __init__(self):

        luttablepath = os.path.join(os.path.dirname(os.path.abspath(__file__)), "metadata", "intentlut.csv")
        lutdf = pd.read_csv(luttablepath, engine = 'python')
        
        counter = 0
        
        self._labeltable = dict()
        
        lutkey = lutdf.columns.tolist()
        for key in lutkey: 
            self._labeltable[key] = counter
            counter += 1
            
        #default out of scope label 
        lutkey['out-of-scope'] = -1
        
    @property
    def labeltable(self):
        return self._labeltable
    
    def filter(self, words : str) -> str:
        # TODO: adaptive change the splitting og strin gaccording to row
        end_index = words.find(" ") 
        return words[:end_index] if end_index != -1 else words

    def pseudolabel(self, words :str) -> int:
        """
        Label a sequence/sentence with intent classification based on heuristics
        
        Attributes
        ----------
        sentence : str
            a single sentence

        Returns
        -------
        int
        0: (Follow Up) Question
        1: (Follow Up) Action
        -1: Out-Of-Scope
        """
        wordsfiltered = self.filter(words)
        
        for key in self._labeltable.keys():
        
            rowsfound = lut.loc[lut[key].eq(targetstr)].shape[0]
        
            if rowsfound != -1:
                return self._labeltable[key]
            
        #default return out of scope label if not found 
        return self._labeltable['out-of-scope']
                
                